from ewah.hooks.base import EWAHBaseHook

import requests
from datetime import datetime, timedelta
from selenium import webdriver


class EWAHPlentyMarketsHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "username": "login",
        "url": "host",
    }

    conn_name_attr = "plentymarkets_conn_id"
    default_conn_name = "plentymarkets_default"
    conn_type = "ewah_plentymarkets"
    hook_name = "EWAH PlentyMarkets Connection"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra"],
            "relabeling": {
                "login": "Username",
                "password": "Password",
                "host": "URL (e.g. https://myplentyshop.com or https://plentymarkets-cloud-07.com/12345)",
            },
        }

    @property
    def endpoint(self):
        # get the current endpoint with the correct backend hash
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--window-size=1420,1080")
        driver = webdriver.Chrome(options=chrome_options)
        driver.get(self.conn.url)
        driver.find_element_by_id("username").send_keys(self.conn.username)
        driver.find_element_by_id("password").send_keys(self.conn.password)
        driver.find_element_by_xpath('//button[normalize-space()="Login"]').click()
        endpoint = driver.current_url
        if endpoint.count("/") > 2:
            endpoint = endpoint[
                : endpoint.find("/", endpoint.find("/", endpoint.find("/") + 1) + 1)
            ]
        return endpoint

    @property
    def token(self):
        if not hasattr(self, "_token") or datetime.now() > self._token_expires_at:
            requested_at = datetime.now()
            token_request = requests.post(
                self.endpoint + "/rest/login",
                params={"username": self.conn.username, "password": self.conn.password},
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
            )
            assert token_request.status_code == 200, token_request.text
            try:
                request_data = token_request.json()
            except:
                assert False, "Response is not a JSON - Response Text: {0}".format(
                    token_request.text
                )
            self._token_expires_at = requested_at + timedelta(
                seconds=request_data["expires_in"]
            )
            self._token = request_data["access_token"]
        return self._token

    def get_data_in_batches(
        self,
        resource,
        data_from=None,
        data_until=None,
        batch_size=10000,
        additional_params=None,
    ):
        params = {
            "itemsPerPage": 250,  # Maximum
        }
        if additional_params:
            assert isinstance(additional_params, dict)
            params.update(additional_params)
        if data_from:
            params["updatedAtFrom"] = data_from.isoformat()
        if data_until:
            params["updatedAtTo"] = data_until.isoformat()
        if "rest/" in resource:
            if not resource.startswith("/"):
                resource = "/" + resource
            url = self.endpoint + resource
        else:
            url = self.endpoint + "/rest/{0}".format(resource)

        data = []
        while True:
            headers = {
                "Accept": "application/json",
                "Content-Type": "application/json",
                "Authorization": "Bearer {0}".format(self.token),
            }
            self.log.info("Requesting new page of data...")
            data_request = requests.get(url, params=params, headers=headers)
            assert data_request.status_code in (200, 401), "Status {0}: {1}".format(
                data_request.status_code, data_request.text
            )
            if data_request.status_code == 401:
                # Sometimes, the backend hash changes. If this happens during
                # a data load, is will result in a specific 401 error. Catch
                # the error and try again (once) with the new endpoint. The
                # try is needed in case the data_request() does not return
                # a JSON (will produce an error otherwise).
                try:
                    if data_request.json()["class"] == "UIHashExpiredException":
                        if "rest/" in resource:
                            url = self.endpoint + resource
                        else:
                            url = self.endpoint + "/rest/{0}".format(resource)
                        data_request = requests.get(url, params=params, headers=headers)
                except:
                    pass  # assert below will take care of any error
            assert data_request.status_code == 200, "Status {0}: {1}".format(
                data_request.status_code, data_request.text
            )
            returned_data = data_request.json()
            data += returned_data["entries"]
            if returned_data["isLastPage"]:
                break
            if len(data) >= batch_size:
                yield data
                data = []
            params["page"] = returned_data["page"] + 1  # for next request
        if data:
            yield data



from airflow import DAG
from airflow.configuration import conf
from airflow.models import Variable
from airflow.operators.bash import BashOperator
from airflow.providers.postgres.operators.postgres import PostgresOperator

from ewah.constants import EWAHConstants as EC
from ewah.ewah_utils.airflow_utils import EWAHSqlSensor, datetime_utcnow_with_tz
from ewah.ewah_utils.dbt_operator import EWAHdbtOperator
from ewah.hooks.base import EWAHBaseHook as BaseHook

from datetime import datetime, timedelta

import os
import pytz


def dbt_dags_factory_legacy(
    dwh_engine,
    dwh_conn_id,
    project_name,
    dbt_schema_name,
    airflow_conn_id,
    dag_base_name="DBT_run",
    analytics_reader=None,  # list of users of DWH who are read-only
    schedule_interval=timedelta(hours=1),
    start_date=datetime(2019, 1, 1),
    default_args=None,
    folder=None,
    models=None,
    exclude=None,
):

    if analytics_reader:
        for statement in (
            "insert",
            "update",
            "delete",
            "drop",
            "create",
            "select",
            ";",
            "grant",
        ):
            for reader in analytics_reader:
                if statement in reader.lower():
                    raise Exception(
                        "Error! The analytics reader {0} "
                        + "is invalid.".format(reader)
                    )

        # analytics_reader = analytics_reader.split(',')
        analytics_reader_sql = f'\nGRANT USAGE ON SCHEMA "{dbt_schema_name}"'
        analytics_reader_sql += ' TO "{0}";'
        analytics_reader_sql += (
            f'''
        \nGRANT SELECT ON ALL TABLES IN SCHEMA "{dbt_schema_name}"'''
            + ' TO "{0}";'
        )
        analytics_reader_sql = "".join(
            [analytics_reader_sql.format(i) for i in analytics_reader]
        )

    if models and not (type(models) == str):
        models = " --models " + " ".join(models)
    else:
        models = ""

    if exclude and not (type(exclude) == str):
        exclude = " --exclude " + " ".join(exclude)
    else:
        exclude = ""

    flags = models + exclude

    dag = DAG(
        dag_base_name,
        catchup=False,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        default_args=default_args,
    )

    dag_full_refresh = DAG(
        dag_base_name + "_full_refresh",
        catchup=False,
        max_active_runs=1,
        schedule_interval=None,
        start_date=start_date,
        default_args=default_args,
    )

    folder = folder or (
        os.environ.get("AIRFLOW_HOME") or conf.get("core", "airflow_home")
    ).replace(
        "airflow_home/airflow",
        "dbt_home",
    )

    bash_command = """
    cd {1}
    source env/bin/activate
    cd {2}
    dbt {0}
    """.format(
        "{0}",
        folder,
        project_name,
    )

    sensor_sql = """
        SELECT
            CASE WHEN COUNT(*) = 0 THEN 1 ELSE 0 END -- only run if exatly equal to 0
        FROM public.dag_run
        WHERE dag_id IN ('{0}', '{1}')
        and state = 'running'
        and not (run_id = '{2}')
    """.format(
        dag._dag_id,
        dag_full_refresh._dag_id,
        "{{ run_id }}",
    )

    # refactor?! not coupled to values in profiles.yml!
    if dwh_engine == EC.DWH_ENGINE_POSTGRES:
        conn = BaseHook.get_connection(dwh_conn_id)
        env = {
            "DBT_DWH_HOST": str(conn.host),
            "DBT_DWH_USER": str(conn.login),
            "DBT_DWH_PASS": str(conn.password),
            "DBT_DWH_PORT": str(conn.port),
            "DBT_DWH_DBNAME": str(conn.schema),
            "DBT_DWH_SCHEMA": dbt_schema_name,
            "DBT_PROFILES_DIR": folder,
        }
    elif dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
        analytics_conn = BaseHook.get_connection(dwh_conn_id)
        analytics_conn_extra = analytics_conn.extra_dejson
        env = {
            "DBT_ACCOUNT": analytics_conn_extra.get(
                "account",
                analytics_conn.host,
            ),
            "DBT_USER": analytics_conn.login,
            "DBT_PASS": analytics_conn.password,
            "DBT_ROLE": analytics_conn_extra.get("role"),
            "DBT_DB": analytics_conn_extra.get("database"),
            "DBT_WH": analytics_conn_extra.get("warehouse"),
            "DBT_SCHEMA": dbt_schema_name,
            "DBT_PROFILES_DIR": folder,
        }
    else:
        raise ValueError("DWH type not implemented!")

    # with dag:
    snsr = EWAHSqlSensor(
        task_id="sense_dbt_conflict_avoided",
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        dag=dag,
    )

    dbt_seed = BashOperator(
        task_id="run_dbt_seed",
        bash_command=bash_command.format("seed"),
        env=env,
        dag=dag,
    )

    dbt_run = BashOperator(
        task_id="run_dbt",
        bash_command=bash_command.format("run" + flags),
        env=env,
        dag=dag,
    )

    dbt_test = BashOperator(
        task_id="test_dbt",
        bash_command=bash_command.format("test" + flags),
        env=env,
        dag=dag,
    )

    dbt_docs = BashOperator(
        task_id="create_dbt_docs",
        bash_command=bash_command.format("docs generate"),
        env=env,
        dag=dag,
    )

    snsr >> dbt_seed >> dbt_run >> dbt_test

    if analytics_reader:
        # This should not occur when using Snowflake
        read_rights = PostgresOperator(
            task_id="grant_access_to_read_users",
            sql=analytics_reader_sql,
            postgres_conn_id=dwh_conn_id,
            dag=dag,
        )
        dbt_test >> read_rights >> dbt_docs
    else:
        dbt_test >> dbt_docs

    # with dag_full_refresh:
    snsr = EWAHSqlSensor(
        task_id="sense_dbt_conflict_avoided",
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        dag=dag_full_refresh,
    )

    dbt_seed = BashOperator(
        task_id="run_dbt_seed",
        bash_command=bash_command.format("seed"),
        env=env,
        dag=dag_full_refresh,
    )

    dbt_run = BashOperator(
        task_id="run_dbt",
        bash_command=bash_command.format("run --full-refresh" + flags),
        env=env,
        dag=dag_full_refresh,
    )

    dbt_test = BashOperator(
        task_id="test_dbt",
        bash_command=bash_command.format("test" + flags),
        env=env,
        dag=dag_full_refresh,
    )

    dbt_docs = BashOperator(
        task_id="create_dbt_docs",
        bash_command=bash_command.format("docs generate"),
        env=env,
        dag=dag_full_refresh,
    )

    snsr >> dbt_seed >> dbt_run >> dbt_test

    if analytics_reader:
        read_rights = PostgresOperator(
            task_id="grant_access_to_read_users",
            sql=analytics_reader_sql,
            postgres_conn_id=dwh_conn_id,
            dag=dag_full_refresh,
        )
        dbt_test >> read_rights >> dbt_docs
    else:
        dbt_test >> dbt_docs
    return (dag, dag_full_refresh)


def dbt_dag_factory_new(
    airflow_conn_id,
    repo_type,
    dwh_engine,
    dwh_conn_id,
    database_name=None,
    git_conn_id=None,  # if provided, expecting private SSH key in conn extra
    dbt_version="0.18.1",
    subfolder=None,  # optional: supply if dbt project is in a subfolder
    threads=4,  # see https://docs.getdbt.com/dbt-cli/configure-your-profile/#understanding-threads
    schema_name="analytics",  # see https://docs.getdbt.com/dbt-cli/configure-your-profile/#understanding-target-schemas
    keepalives_idle=0,  # see https://docs.getdbt.com/reference/warehouse-profiles/postgres-profile/
    dag_base_name="T_dbt_run",
    schedule_interval=timedelta(hours=1),
    start_date=datetime(2019, 1, 1),
    default_args=None,
    run_flags=None,  # e.g. --model tag:base
    project=None,  # BigQuery alias
    dataset=None,  # BigQuery alias
    dagrun_timeout_factor=0.8,
):
    run_flags = run_flags or ""  # use empty string instead of None

    # only PostgreSQL & Snowflake implemented as of now!
    assert dwh_engine in (
        EC.DWH_ENGINE_POSTGRES,
        EC.DWH_ENGINE_SNOWFLAKE,
        EC.DWH_ENGINE_BIGQUERY,
    )

    # if start_date is timezone offset-naive, assume utc and turn into offset-aware
    if not start_date.tzinfo:
        start_date = start_date.replace(tzinfo=pytz.utc)

    start_date += (
        int((datetime_utcnow_with_tz() - start_date) / schedule_interval) - 1
    ) * schedule_interval
    end_date = start_date + 2 * schedule_interval - timedelta(seconds=1)

    dag_kwargs = {
        "catchup": True,
        "start_date": start_date,
        "end_date": end_date,
        "default_args": default_args,
        "max_active_runs": 1,
    }

    if dagrun_timeout_factor:
        _msg = "dagrun_timeout_factor must be a number between 0 and 1!"
        assert isinstance(dagrun_timeout_factor, (int, float)) and (
            0 < dagrun_timeout_factor <= 1
        ), _msg
        dag_kwargs["dagrun_timeout"] = dagrun_timeout_factor * schedule_interval
        dag_kwargs["default_args"]["execution_timeout"] = dag_kwargs[
            "default_args"
        ].get("execution_timeout", dag_kwargs["dagrun_timeout"])

    dag_1 = DAG(dag_base_name, schedule_interval=schedule_interval, **dag_kwargs)
    dag_2 = DAG(dag_base_name + "_full_refresh", schedule_interval=None, **dag_kwargs)

    sensor_sql = """
        SELECT
            -- only run if exactly equal to 0
            CASE WHEN COUNT(*) = 0 THEN 1 ELSE 0 END
        FROM public.dag_run
        WHERE dag_id IN ('{0}', '{1}')
          AND state = 'running'
          AND not (run_id = '{2}')
          AND (( -- Avoid deadlocks and prioritize scheduled over manual triggers
              run_type = 'scheduled'
              AND execution_date < '{3}' -- execution_date
            ) OR (
              run_type = 'manual'
              AND execution_date < '{4}' -- next_execution_date
          ))
          -- Note: next_execution_date = execution_date if run_type = 'manual'
    """.format(
        dag_1._dag_id,
        dag_2._dag_id,
        "{{ run_id }}",
        "{{ execution_date }}",
        "{{ next_execution_date }}",
    )

    snsr_1 = EWAHSqlSensor(
        task_id="sense_dbt_conflict_avoided",
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        dag=dag_1,
    )
    snsr_2 = EWAHSqlSensor(
        task_id="sense_dbt_conflict_avoided",
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        dag=dag_2,
    )

    dbt_kwargs = {
        "repo_type": repo_type,
        "dwh_conn_id": dwh_conn_id,
        "git_conn_id": git_conn_id,
        "dbt_version": dbt_version,
        "subfolder": subfolder,
        "threads": threads,
        "schema_name": schema_name,
        "keepalives_idle": 0,
        "dwh_engine": dwh_engine,
        "database_name": database_name,
        "project": project,
        "dataset": dataset,
    }

    run_1 = EWAHdbtOperator(
        task_id="dbt_run",
        dbt_commands=["seed", f"run {run_flags}"],
        dag=dag_1,
        **dbt_kwargs,
    )
    run_2 = EWAHdbtOperator(
        task_id="dbt_run",
        dbt_commands=["seed --full-refresh", f"run --full-refresh {run_flags}"],
        dag=dag_2,
        **dbt_kwargs,
    )

    test_1 = EWAHdbtOperator(
        task_id="dbt_test", dbt_commands="test", dag=dag_1, **dbt_kwargs
    )
    test_2 = EWAHdbtOperator(
        task_id="dbt_test", dbt_commands="test", dag=dag_2, **dbt_kwargs
    )

    snsr_1 >> run_1 >> test_1
    snsr_2 >> run_2 >> test_2

    return (dag_1, dag_2)


def dbt_dags_factory(*args, **kwargs):
    if "dbt_schema_name" in kwargs:
        return dbt_dags_factory_legacy(*args, **kwargs)
    else:
        return dbt_dag_factory_new(*args, **kwargs)


def dbt_snapshot_dag(
    dwh_engine,
    dwh_conn_id,
    git_conn_id,
    database_name=None,
    dbt_version="0.18.1",
    subfolder=None,
    threads=4,
    schema_name="analytics",  # for the profiles.yml
    keepalives_idle=0,
    dag_name="T_dbt_snapshots",
    schedule_interval=timedelta(hours=1),
    start_date=None,
    default_args=None,
    dagrun_timeout_factor=0.8,
):
    # only PostgreSQL & Snowflake implemented as of now!
    assert dwh_engine in (EC.DWH_ENGINE_POSTGRES, EC.DWH_ENGINE_SNOWFLAKE)

    if dagrun_timeout_factor:
        _msg = "dagrun_timeout_factor must be a number between 0 and 1!"
        assert isinstance(dagrun_timeout_factor, (int, float)) and (
            0 < dagrun_timeout_factor <= 1
        ), _msg
        dagrun_timeout = dagrun_timeout_factor * schedule_interval
        default_args = default_args or {}
        default_args["execution_timeout"] = default_args.get(
            "execution_timeout", dagrun_timeout
        )
    else:  # In case of 0 set to None
        dagrun_timeout = None

    dag = DAG(
        dag_name,
        schedule_interval=schedule_interval,
        catchup=False,
        max_active_runs=1,
        start_date=start_date,
        default_args=default_args,
        dagrun_timeout=dagrun_timeout,
    )

    task = EWAHdbtOperator(
        dag=dag,
        task_id="dbt_snapshot",
        dbt_commands=["snapshot"],
        repo_type="git",
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        git_conn_id=git_conn_id,
        database_name=database_name,
        dbt_version=dbt_version,
        subfolder=subfolder,
        threads=threads,
        schema_name=schema_name,
        keepalives_idle=keepalives_idle,
    )

    return dag



from ewah.constants import EWAHConstants as EC
from ewah.hooks.plentymarkets import EWAHPlentyMarketsHook
from ewah.operators.base import EWAHBaseOperator

"""temporary notes

- use selenium with phantomjs (https://phantomjs.org/download.html)

- find button via normalize space: https://stackoverflow.com/questions/49906237/how-to-find-button-with-selenium-by-its-text-inside-python

d = webdriver.PhantomJS("c:/phantomjs-2.1.1-windows/bin/phantomjs.exe")
d.set_window_size(1120, 550)
d.get("https://plentymarkets-cloud-07.com/50366")
d.find_element_by_id("username").send_keys("ga_bijan.soltani")
d.find_element_by_id("password").send_keys("jeak6KOSH6pird!waff")
d.find_element_by_xpath('//button[normalize-space()="Login"]').click()
d.current_url
"""


class EWAHPlentyMarketsOperator(EWAHBaseOperator):

    _NAMES = ["plentymarkets"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHPlentyMarketsHook.conn_type

    _REQUIRES_COLUMNS_DEFINITION = False

    def __init__(self, resource=None, additional_api_call_params=None, *args, **kwargs):
        kwargs["primary_key_column_name"] = "id"
        resource = resource or kwargs.get("target_table_name")
        if kwargs["extract_strategy"] == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = kwargs.get("subsequent_field", "updatedAt")
            # currently, only the orders resource works with subsequent loading
            assert resource == "orders"
        if kwargs["extract_strategy"] == EC.ES_INCREMENTAL:
            # currently, only the orders resource works with incremental loading
            assert resource == "orders"
        super().__init__(*args, **kwargs)

        assert isinstance(additional_api_call_params, (type(None), dict))
        self.resource = resource
        self.additional_api_call_params = additional_api_call_params

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
        else:
            data_from = self.data_from
        for batch in self.source_hook.get_data_in_batches(
            resource=self.resource,
            data_from=data_from,
            data_until=self.data_until,
            additional_params=self.additional_api_call_params,
        ):
            self.upload_data(batch)



from ewah.constants import EWAHConstants as EC
from ewah.hooks.base import EWAHBaseHook

from apiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials as SAC

import json
import time
from datetime import datetime, timedelta


class EWAHGoogleAnalyticsHook(EWAHBaseHook):

    _ATTR_RELABEL = {}

    conn_name_attr = "ewah_google_analytics_conn_id"
    default_conn_name = "ewah_google_analytics_default"
    conn_type = "ewah_google_analytics"
    hook_name = "EWAH Google Analytics Connection"

    METRICMAP = {
        "METRIC_TYPE_UNSPECIFIED": "varchar(255)",
        "CURRENCY": "decimal(20,5)",
        "INTEGER": "int(11)",
        "FLOAT": "decimal(20,5)",
        "PERCENT": "decimal(20,5)",
        "TIME": "time",
    }

    _SAMPLE_JSON = {
        "client_secrets": {
            "type": "service_account",
            "project_id": "abc-123",
            "private_key_id": "123456abcder",
            "private_key": "-----BEGIN PRIVATE KEY-----\nxxx\n-----END PRIVATE KEY-----\n",
            "client_email": "xyz@abc-123.iam.gserviceaccount.com",
            "client_id": "123457",
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/xyz%40abc-123.iam.gserviceaccount.com",
        }
    }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra", "password", "login", "schema", "host", "port"],
            "relabeling": {},
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        # from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from ewah.ewah_utils.widgets import EWAHTextAreaWidget
        from wtforms import StringField

        return {
            "extra__ewah_google_analytics__service_account_json": StringField(
                "Service Account JSON",
                widget=EWAHTextAreaWidget(rows=12),
            ),
        }

    def get_data_in_batches(self, data_from, data_until, chunking_interval, **kwargs):
        while data_from <= data_until:
            yield self.get_data(
                data_from=data_from,
                data_until=min(data_until, data_from + chunking_interval),
                **kwargs
            )
            data_from += chunking_interval + timedelta(days=1)

    def get_data(
        self,
        view_id,
        dimensions,
        metrics,
        page_size,
        include_empty_rows,
        sampling_level,
        data_from,
        data_until,
    ):
        sac_json = self.conn.service_account_json
        try:
            sac_json = json.loads(sac_json)
        except:
            raise Exception("Service Account JSON is not valid JSON!")

        if sac_json.get("client_secrets"):
            sac_json = sac_json.get("client_secrets")

        if not all(
            [key in sac_json.keys() for key in self._SAMPLE_JSON["client_secrets"]]
        ) and len(sac_json.keys()) == len(self._SAMPLE_JSON["client_secrets"]):
            _msg = "Google Analytics Credentials misspecified!"
            _msg += " Example of a correct Service Account JSON: {0}".format(
                json.dumps(self._SAMPLE_JSON)
            )
            raise Exception(_msg)

        if len(dimensions) > 7:
            raise Exception(
                (
                    "Can only fetch up to 7 dimensions!" + " Currently {0} Dimensions"
                ).format(
                    str(len(dimensions)),
                )
            )

        if len(metrics) > 10:
            raise Exception(
                (
                    "Can only fetch up to 10 metrics!" + " Currently {0} Dimensions"
                ).format(
                    str(len(metrics)),
                )
            )

        if page_size > 10000:
            raise Exception("Please specify a page size equal to or lower than 10000.")

        service_object = build(
            "analyticsreporting",
            "v4",
            credentials=SAC.from_json_keyfile_dict(
                sac_json,
                ["https://www.googleapis.com/auth/analytics.readonly"],
            ),
        )

        report_request = {
            "viewId": view_id,
            "dateRanges": [
                {
                    "startDate": data_from.isoformat(),
                    "endDate": data_until.isoformat(),
                }
            ],
            "samplingLevel": sampling_level,
            "dimensions": [{"name": d} for d in dimensions],
            "metrics": [{"expression": m} for m in metrics],
            "pageSize": page_size,
            "includeEmptyRows": include_empty_rows,
        }

        self.log.info(
            "Loading data from {0} to {1}...".format(
                report_request["dateRanges"][0]["startDate"],
                report_request["dateRanges"][0]["endDate"],
            )
        )

        report_response = (
            service_object.reports()
            .batchGet(body={"reportRequests": [report_request]})
            .execute()
        )

        if report_response.get("reports"):
            report = report_response["reports"][0]
            rows = report.get("data", {}).get("rows", [])

            while report.get("nextPageToken"):
                time.sleep(1)
                report_request.update({"pageToken": report["nextPageToken"]})
                report_response = (
                    service_object.reports()
                    .batchGet(body={"reportRequests": [report_request]})
                    .execute()
                )
                report = report_response["reports"][0]
                rows.extend(report.get("data", {}).get("rows", []))

            if report["data"]:
                report["data"]["rows"] = rows
        else:
            report = {}

        column_header = report.get("columnHeader", {})
        # Right now all dimensions are hardcoded to varchar(255), will need a
        # map if any non-varchar dimensions are used in the future
        # Unfortunately the API does not send back types for Dimensions like it
        # does for Metrics (yet..)
        dimension_headers = [
            {
                "name": header.replace("ga:", ""),
                "type": "varchar(255)",
            }
            for header in column_header.get("dimensions", [])
        ]
        metric_headers = [
            {
                "name": entry.get("name").replace("ga:", ""),
                "type": self.METRICMAP.get(entry.get("type"), "varchar(255)"),
            }
            for entry in column_header.get("metricHeader", {}).get(
                "metricHeaderEntries", []
            )
        ]

        self.log.info("Reading data from fetched report...")
        uploadable_data = []
        rows = report.get("data", {}).get("rows", [])
        for row_counter, row in enumerate(rows):
            root_data_obj = {}
            dimensions = row.get("dimensions", [])
            metrics = row.get("metrics", [])

            for index, dimension in enumerate(dimensions):
                header = dimension_headers[index].get("name")  # .lower()
                root_data_obj[header] = dimension

            for metric in metrics:
                data = {}
                data.update(root_data_obj)

                for index, value in enumerate(metric.get("values", [])):
                    header = metric_headers[index].get("name")  # .lower()
                    data[header] = value

                data["view_id"] = view_id
                if data.get("date"):
                    data["date"] = datetime.strptime(data["date"], "%Y%m%d").date()
                if data.get("dateHour"):
                    data["dateHour"] = datetime.strptime(data["dateHour"], "%Y%m%d%H")

                uploadable_data += [data]

        return uploadable_data



from ewah.hooks.base import EWAHBaseHook
import recurly


class EWAHRecurlyHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "api_token": "password",
    }

    conn_name_attr = "recurly_conn_id"
    default_conn_name = "recurly_default"
    conn_type = "ewah_recurly"
    hook_name = "EWAH Recurly Connection"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra", "host", "login"],
            "relabeling": {"password": "Baisc Auth API Key"},
        }

    @property
    def client(self):
        if not hasattr(self, "_client"):
            self._client = recurly.Client(self.conn.api_token)

        return self._client

    @staticmethod
    def validate_resource(resource):
        return hasattr(recurly.Client, "list_{0}".format(resource))

    @classmethod
    def recurly_to_dict(cls, recurly_object):
        final_dict = {}
        for key, value in vars(recurly_object).items():
            if isinstance(value, recurly.Resource):
                value = cls.recurly_to_dict(value)
            if isinstance(value, list):
                value = [
                    cls.recurly_to_dict(item)
                    if isinstance(item, recurly.Resource)
                    else item
                    for item in value
                ]
            final_dict[key] = value
        return final_dict

    def get_data_in_batches(
        self,
        resource,
        data_from=None,
        data_until=None,
        batch_size=10000,
    ):
        assert isinstance(batch_size, int) and batch_size > 0
        params = {
            "limit": 200,  # maximum limit per call (overwrite default of 20)
            "sort": "updated_at",
            "order": "asc",  # see docs - don't use desc with updated_at sort!
        }
        if data_from:
            params["begin_time"] = data_from.isoformat()
        if data_until:
            params["end_time"] = data_until.isoformat()

        # Serve data in batches
        data = []
        i = 0
        for item in getattr(self.client, "list_{0}".format(resource))(
            params=params
        ).items():
            i += 1
            data.append(self.recurly_to_dict(item))
            if i == batch_size:
                yield data
                data = []
                i = 0

        if data:  # last batch
            yield data



from ewah.hooks.base import EWAHBaseHook

from typing import Optional, List, Dict, Any
from collections import defaultdict

import boto3


class EWAHAWSHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "access_key_id": "login",
        "secret_access_key": "password",
        "region": "schema",
    }

    conn_name_attr = "ewah_aws_conn_id"
    default_conn_name = "ewah_aws_default"
    conn_type = "ewah_aws"
    hook_name = "EWAH AWS Connection"

    _resources = defaultdict(dict)

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "extra", "host"],
            "relabeling": {
                "login": "AWS Access Key ID",
                "password": "AWS Secret Access Key",
                "schema": "Region",
            },
        }

    def get_boto_resource(self, resource: str, region: Optional[str] = None):
        aws_region = region or self.region
        if not self._resources[aws_region].get(resource):
            self._resources[aws_region][resource] = boto3.resource(
                resource,
                aws_access_key_id=self.conn.access_key_id,
                aws_secret_access_key=self.conn.secret_access_key,
                region_name=aws_region,
            )
        return self._resources[aws_region][resource]

    def get_dynamodb_data_in_batches(
        self,
        table_name: str,
        region: Optional[str] = None,
        batch_size: int = 10000,
        pagination_limit: Optional[int] = None,
        filter_expression=None,
    ) -> List[Dict[str, Any]]:
        resource = self.get_boto_resource(resource="dynamodb", region=region)
        table = resource.Table(table_name)

        # build scan kwargs
        scan_kwargs = {}
        if pagination_limit:
            scan_kwargs["Limit"] = pagination_limit
        if filter_expression:
            scan_kwargs["FilterExpression"] = filter_expression

        # iterate through entire table
        keepgoing = True
        batch_data = []
        while keepgoing:
            response = table.scan(**scan_kwargs)
            batch_data += response.get("Items")
            scan_kwargs["ExclusiveStartKey"] = response.get("LastEvaluatedKey")
            keepgoing = bool(scan_kwargs["ExclusiveStartKey"])
            if len(batch_data) >= batch_size or not keepgoing:
                yield batch_data
                batch_data = []



from airflow.operators.python import PythonOperator as PO
from airflow.operators.dummy import DummyOperator as DO
from airflow.models import BaseOperator
from airflow.sensors.sql import SqlSensor

from ewah.hooks.base import EWAHBaseHook
from ewah.constants import EWAHConstants as EC

from datetime import datetime, timedelta, timezone
from copy import deepcopy
import pytz
import re


class EWAHSqlSensor(SqlSensor):
    """Overwrite native SQL sensor to allow usage of ewah custom connection types."""

    def _get_hook(self):
        conn = EWAHBaseHook.get_connection(conn_id=self.conn_id)
        if not conn.conn_type.startswith("ewah"):
            raise Exception("Must use an appropriate EWAH custom connection type!")
        return conn.get_hook()


class PGO(BaseOperator):
    """Airflow operator to execute PostgreSQL statements.

    Cannot use airflow.providers.postgres.operators.postgres.PostgresOperator
        due to the a requirement to have an SSH tunnel option.

    Needs to be otherwise interchangeable with Airflow's PostgresOperator.
    """

    template_fields = ("sql",)
    template_ext = (".sql",)

    def __init__(self, sql, postgres_conn_id, parameters=None, *args, **kwargs):
        self.sql = sql
        self.postgres_conn_id = postgres_conn_id
        self.parameters = parameters
        super().__init__(*args, **kwargs)

    def execute(self, context):
        hook = EWAHBaseHook.get_hook_from_conn_id(self.postgres_conn_id)
        hook.execute(sql=self.sql, params=self.parameters, commit=True)
        hook.close()  # SSH tunnel does not close if hook is not closed first


def datetime_utcnow_with_tz():
    return datetime.utcnow().replace(tzinfo=pytz.utc)


def airflow_datetime_adjustments(datetime_raw):
    if type(datetime_raw) == str:
        datetime_string = datetime_raw
        if "-" in datetime_string[10:]:
            tz_sign = "-"
        else:
            tz_sign = "+"
        datetime_strings = datetime_string.split(tz_sign)

        if "T" in datetime_string:
            format_string = "%Y-%m-%dT%H:%M:%S"
        else:
            format_string = "%Y-%m-%d %H:%M:%S"

        if "." in datetime_string:
            format_string += ".%f"

        if len(datetime_strings) == 2:
            if ":" in datetime_strings[1]:
                datetime_strings[1] = datetime_strings[1].replace(":", "")
            datetime_string = tz_sign.join(datetime_strings)
            format_string += "%z"
        elif "Z" in datetime_string:
            format_string += "Z"
        datetime_raw = datetime.strptime(
            datetime_string,
            format_string,
        )
    elif not (datetime_raw is None or isinstance(datetime_raw, datetime)):
        raise Exception(
            "Invalid datetime type supplied! Supply either string"
            + " or datetime.datetime! supplied: {0}".format(str(type(datetime_raw)))
        )

    if datetime_raw and datetime_raw.tzinfo is None:
        datetime_raw = datetime_raw.replace(tzinfo=timezone.utc)

    return datetime_raw


def etl_schema_tasks(
    dag,
    dwh_engine,
    dwh_conn_id,
    target_schema_name,
    target_schema_suffix="_next",
    target_database_name=None,
    read_right_users=None,  # Only for PostgreSQL
    **additional_task_args
):

    if dwh_engine == EC.DWH_ENGINE_POSTGRES:
        sql_kickoff = """
            DROP SCHEMA IF EXISTS "{schema_name}{schema_suffix}" CASCADE;
            CREATE SCHEMA "{schema_name}{schema_suffix}";
        """.format(
            schema_name=target_schema_name,
            schema_suffix=target_schema_suffix,
        )
        sql_final = """
            DROP SCHEMA IF EXISTS "{schema_name}" CASCADE;
            ALTER SCHEMA "{schema_name}{schema_suffix}"
                RENAME TO "{schema_name}";
        """.format(
            schema_name=target_schema_name,
            schema_suffix=target_schema_suffix,
        )

        # Don't fail final task just because a user or role that should
        # be granted read rights does not exist!
        grant_rights_sql = """
            DO $$
            BEGIN
              GRANT USAGE ON SCHEMA "{target_schema_name}" TO {user};
              GRANT SELECT ON ALL TABLES
                IN SCHEMA "{target_schema_name}" TO {user};
              EXCEPTION WHEN OTHERS THEN -- catches any error
                RAISE NOTICE 'not granting rights - user does not exist!';
            END
            $$;
        """
        if read_right_users:
            if not isinstance(read_right_users, list):
                raise Exception("Arg read_right_users must be of type List!")
            for user in read_right_users:
                if re.search(r"\s", user) or (";" in user):
                    _msg = "No whitespace or semicolons allowed in usernames!"
                    raise ValueError(_msg)
                sql_final += grant_rights_sql.format(
                    target_schema_name=target_schema_name,
                    user=user,
                )

        task_1_args = deepcopy(additional_task_args)
        task_2_args = deepcopy(additional_task_args)
        task_1_args.update(
            {
                "sql": sql_kickoff,
                "task_id": "kickoff_{0}".format(target_schema_name),
                "dag": dag,
                "postgres_conn_id": dwh_conn_id,
            }
        )
        task_2_args.update(
            {
                "sql": sql_final,
                "task_id": "final_{0}".format(target_schema_name),
                "dag": dag,
                "postgres_conn_id": dwh_conn_id,
            }
        )
        return (PGO(**task_1_args), PGO(**task_2_args))
    elif dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
        target_database_name = target_database_name or (
            EWAHBaseHook.get_connection(dwh_conn_id).database
        )
        sql_kickoff = """
            DROP SCHEMA IF EXISTS
                "{database}"."{schema_name}{schema_suffix}" CASCADE;
            CREATE SCHEMA "{database}"."{schema_name}{schema_suffix}";
        """.format(
            database=target_database_name,
            schema_name=target_schema_name,
            schema_suffix=target_schema_suffix,
        )
        sql_final = """
            DROP SCHEMA IF EXISTS "{database}"."{schema_name}" CASCADE;
            ALTER SCHEMA "{database}"."{schema_name}{schema_suffix}"
                RENAME TO "{schema_name}";
        """.format(
            database=target_database_name,
            schema_name=target_schema_name,
            schema_suffix=target_schema_suffix,
        )

        def execute_snowflake(sql, conn_id, **kwargs):
            hook = EWAHBaseHook.get_hook_from_conn_id(conn_id)
            hook.execute(sql)
            hook.commit()
            hook.close()

        task_1_args = deepcopy(additional_task_args)
        task_2_args = deepcopy(additional_task_args)
        task_1_args.update(
            {
                "task_id": "kickoff_{0}".format(target_schema_name),
                "python_callable": execute_snowflake,
                "op_kwargs": {
                    "sql": sql_kickoff,
                    "conn_id": dwh_conn_id,
                },
                "provide_context": True,
                "dag": dag,
            }
        )
        task_2_args.update(
            {
                "task_id": "final_{0}".format(target_schema_name),
                "python_callable": execute_snowflake,
                "op_kwargs": {
                    "sql": sql_final,
                    "conn_id": dwh_conn_id,
                },
                "provide_context": True,
                "dag": dag,
            }
        )
        return (PO(**task_1_args), PO(**task_2_args))
    elif dwh_engine == EC.DWH_ENGINE_GS:
        # create dummy tasks
        return (
            DO(
                task_id="kickoff",
                dag=dag,
            ),
            DO(
                task_id="final",
                dag=dag,
            ),
        )
    else:
        raise ValueError("Feature not implemented!")



from ewah.operators.base import EWAHBaseOperator
from ewah.ewah_utils.python_utils import is_iterable_not_string
from ewah.constants import EWAHConstants as EC

from google.ads.google_ads.client import GoogleAdsClient
from ewah.hooks.base import EWAHBaseHook as BaseHook

from datetime import datetime, timedelta
from copy import deepcopy


class EWAHGoogleAdsOperator(EWAHBaseOperator):

    _NAMES = ["gads", "google_ads"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: True,
    }

    _REQUIRED_KEYS = (
        # You must have a developer_token to use the Google Ads API!
        "developer_token",
        "client_id",
        "client_secret",
        "refresh_token",
    )

    def __init__(
        self,
        fields,  # dict of fields, e.g. {'campaign': ['id', 'name']}
        metrics,  # list of metrics to retrieve -
        resource,  # name of the resource, e.g. ad_group
        client_id,
        conditions=None,  # list of conditions, e.g. ["ad_group.status = 'ENABLED'"]
        *args,
        **kwargs
    ):
        """
        recommendation: leave data_until=None and use a timedelta for
        data_from.
        """

        if kwargs.get("columns_definition"):
            raise Exception(
                "columns_definition is not accepted for this " + "operator!"
            )

        kwargs["update_on_columns"] = [
            col.replace(".", "_")
            for col in self.get_select_statement(fields)
            if col[:7] == "segment"
        ] + ["{0}_resource_name".format(resource)]

        if conditions and not is_iterable_not_string(conditions):
            raise Exception('Argument "conditions" must be a list!')

        if not (fields.get("segments") and ("date" in fields["segments"])):
            raise Exception("fields list MUST contain segments.date!")

        # must use deepcopy due to airflow calling field by reference
        # and the update in the following row adding the metrics to the field,
        # which in turn screws with the update_on_columns definition above
        self.fields_dict = deepcopy(fields)
        self.fields_dict.update({"metrics": deepcopy(metrics)})
        self.fields_list = self.get_select_statement(self.fields_dict)
        self.resource = resource
        self.client_id = str(client_id)
        self.conditions = conditions

        super().__init__(*args, **kwargs)

    def get_select_statement(self, dict_format, prefix=None):
        # create the list of fields for the SELECT statement
        if prefix is None:
            prefix = ""
        elif not prefix[-1] == ".":
            prefix += "."
        fields = []
        for key, value in dict_format.items():
            for item in value:
                if type(item) == dict:
                    fields += self.get_select_statement(item, prefix + key)
                else:
                    fields += [prefix + key + "." + item]
        return fields

    def ewah_execute(self, context):
        # Task execution happens here
        def get_data_from_ads_output(fields_dict, values, prefix=None):
            if prefix is None:
                prefix = ""
            elif not prefix[-1] == "_":
                prefix += "_"
                # e.g. 2b prefix = 'ad_group_criterion_'
            data = {}
            for key, value in fields_dict.items():
                # e.g. 1 key = 'metrics', value = ['impressions', 'clicks']
                # e.g. 2a key = 'ad_group_criterion', value = [{'keyword': ['text', 'match_type']}]
                # e.g. 2b key = 'keyword', value = ['text', 'match_type']
                node = getattr(values, key)
                # e.g. 1 node = row.metrics
                # e.g. 2a node = row.ad_group_criterion
                # e.g. 2b node = row.ad_group_criterion.keyword
                for item in value:
                    # e.g. 1 item = 'clicks'
                    # e.g. 2a item = {'keyword': ['text', 'match_type']}
                    # e.g. 2b item = 'text'
                    if type(item) == dict:
                        data.update(
                            get_data_from_ads_output(
                                fields_dict=item,
                                values=node,
                                prefix=prefix
                                + key,  # e.g. 2a '' + 'ad_group_criterion'
                            )
                        )
                    else:
                        # e.g. 1: {'' + 'metrics' + '_' + 'clicks': row.metrics.clicks.value}
                        # e.g. 2b: {'ad_group_criterion_' + 'keyeword' + '_' + 'text': row.ad_group_criterion.keyword.text.value}
                        if hasattr(getattr(node, item), "value"):
                            data.update(
                                {prefix + key + "_" + item: getattr(node, item).value}
                            )
                        else:
                            # some node ends don't respond to .value but are
                            #   already the value
                            data.update(
                                {prefix + key + "_" + item: getattr(node, item)}
                            )
            return data

        conn = self.source_conn.extra_dejson
        credentials = {}
        for key in self._REQUIRED_KEYS:
            if not key in conn.keys():
                raise Exception("{0} must be in connection extra json!".format(key))
            credentials[key] = conn[key]

        # build the query
        query = "SELECT {0} FROM {1} WHERE segments.date {2} {3}".format(
            ", ".join(self.fields_list),
            self.resource,
            "BETWEEN '{0}' AND '{1}'".format(
                self.data_from.strftime("%Y-%m-%d"),
                self.data_until.strftime("%Y-%m-%d"),
            ),
            ("AND" + " AND ".join(self.conditions)) if self.conditions else "",
        )

        self.log.info("executing this google ads query:\n{0}".format(query))
        cli = GoogleAdsClient.load_from_dict(credentials)
        service = cli.get_service("GoogleAdsService")
        search = service.search(
            self.client_id.replace("-", ""),
            query=query,
        )
        data = [row for row in search]

        # get into uploadable format
        upload_data = []
        fields_dict = deepcopy(self.fields_dict)
        fields_dict.update({self.resource: ["resource_name"]})
        while data:
            datum = data.pop(0)
            upload_data += [
                get_data_from_ads_output(
                    fields_dict,
                    datum,
                )
            ]

        self.upload_data(upload_data)



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC
from ewah.hooks.salesforce import EWAHSalesforceHook

from typing import Optional


class EWAHSalesforceOperator(EWAHBaseOperator):

    _NAMES = ["sf", "salesforce"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    _CONN_TYPE = EWAHSalesforceHook.conn_type

    def __init__(
        self, salesforce_object: Optional[str] = None, *args, **kwargs
    ) -> None:
        self.salesforce_object = salesforce_object or kwargs.get("target_table_name")
        kwargs["primary_key_column_name"] = "Id"
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context: dict) -> None:
        self.log.info(f"Fetching Salesforce data for {self.salesforce_object}...")
        for batch in self.source_hook.get_data_in_batches(
            salesforce_object=self.salesforce_object,
            columns=list(self.columns_definition or []) or None,
            data_from=self.data_from,
            data_until=self.data_until,
        ):
            self.upload_data(batch)
